#ifndef INCLUDE_ONCE_B6BD8F6C_3C44_4198_92BD_6170085914FF
#define INCLUDE_ONCE_B6BD8F6C_3C44_4198_92BD_6170085914FF

constexpr double astronomicalUnit=149'597'870'700.; // m
constexpr double AU=astronomicalUnit;

constexpr double sunRadius=696350e3; /* m */
constexpr auto moonRadius=1737.1e3; /* m */

#endif
