#pragma once

extern const char*const PROJECT_VERSION;
#define DATA_ROOT_DIR "/usr/share/CalcMySky/"
#define INSTALL_BINDIR "/usr/bin/"
#define BUILD_BINDIR "/build/calcmysky/src/build/"
#define SOURCE_DIR "/build/calcmysky/src/CalcMySky/"
