/****************************************************************************
** Meta object code from reading C++ file 'ToolsWidget.hpp'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../CalcMySky/ShowMySky/ToolsWidget.hpp"
#include <QtGui/qtextcursor.h>
#include <QtGui/qscreen.h>
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'ToolsWidget.hpp' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN11ToolsWidgetE_t {};
} // unnamed namespace

template <> constexpr inline auto ToolsWidget::qt_create_metaobjectdata<qt_meta_tag_ZN11ToolsWidgetE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "ToolsWidget",
        "settingChanged",
        "",
        "ditheringMethodChanged",
        "setScattererEnabled",
        "name",
        "enable",
        "reloadShadersClicked",
        "setFlatSolarSpectrum",
        "resetSolarSpectrum",
        "setBlackBodySolarSpectrum",
        "temperature",
        "windowDecorationToggled",
        "enabled",
        "projectionChanged",
        "GLWidget::Projection",
        "colorModeChanged",
        "GLWidget::ColorMode"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'settingChanged'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'ditheringMethodChanged'
        QtMocHelpers::SignalData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'setScattererEnabled'
        QtMocHelpers::SignalData<void(QString const &, bool)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 5 }, { QMetaType::Bool, 6 },
        }}),
        // Signal 'reloadShadersClicked'
        QtMocHelpers::SignalData<void()>(7, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'setFlatSolarSpectrum'
        QtMocHelpers::SignalData<void()>(8, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'resetSolarSpectrum'
        QtMocHelpers::SignalData<void()>(9, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'setBlackBodySolarSpectrum'
        QtMocHelpers::SignalData<void(double)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Double, 11 },
        }}),
        // Signal 'windowDecorationToggled'
        QtMocHelpers::SignalData<void(bool)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 13 },
        }}),
        // Signal 'projectionChanged'
        QtMocHelpers::SignalData<void(GLWidget::Projection)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 15, 2 },
        }}),
        // Signal 'colorModeChanged'
        QtMocHelpers::SignalData<void(GLWidget::ColorMode)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 17, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<ToolsWidget, qt_meta_tag_ZN11ToolsWidgetE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject ToolsWidget::staticMetaObject = { {
    QMetaObject::SuperData::link<QDockWidget::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN11ToolsWidgetE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN11ToolsWidgetE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN11ToolsWidgetE_t>.metaTypes,
    nullptr
} };

void ToolsWidget::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<ToolsWidget *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->settingChanged(); break;
        case 1: _t->ditheringMethodChanged(); break;
        case 2: _t->setScattererEnabled((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 3: _t->reloadShadersClicked(); break;
        case 4: _t->setFlatSolarSpectrum(); break;
        case 5: _t->resetSolarSpectrum(); break;
        case 6: _t->setBlackBodySolarSpectrum((*reinterpret_cast<std::add_pointer_t<double>>(_a[1]))); break;
        case 7: _t->windowDecorationToggled((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 8: _t->projectionChanged((*reinterpret_cast<std::add_pointer_t<GLWidget::Projection>>(_a[1]))); break;
        case 9: _t->colorModeChanged((*reinterpret_cast<std::add_pointer_t<GLWidget::ColorMode>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (ToolsWidget::*)()>(_a, &ToolsWidget::settingChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (ToolsWidget::*)()>(_a, &ToolsWidget::ditheringMethodChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (ToolsWidget::*)(QString const & , bool )>(_a, &ToolsWidget::setScattererEnabled, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (ToolsWidget::*)()>(_a, &ToolsWidget::reloadShadersClicked, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (ToolsWidget::*)()>(_a, &ToolsWidget::setFlatSolarSpectrum, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (ToolsWidget::*)()>(_a, &ToolsWidget::resetSolarSpectrum, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (ToolsWidget::*)(double )>(_a, &ToolsWidget::setBlackBodySolarSpectrum, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (ToolsWidget::*)(bool )>(_a, &ToolsWidget::windowDecorationToggled, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (ToolsWidget::*)(GLWidget::Projection )>(_a, &ToolsWidget::projectionChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (ToolsWidget::*)(GLWidget::ColorMode )>(_a, &ToolsWidget::colorModeChanged, 9))
            return;
    }
}

const QMetaObject *ToolsWidget::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *ToolsWidget::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN11ToolsWidgetE_t>.strings))
        return static_cast<void*>(this);
    if (!strcmp(_clname, "ShowMySky::Settings"))
        return static_cast< ShowMySky::Settings*>(this);
    return QDockWidget::qt_metacast(_clname);
}

int ToolsWidget::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QDockWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 10)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 10;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 10)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 10;
    }
    return _id;
}

// SIGNAL 0
void ToolsWidget::settingChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void ToolsWidget::ditheringMethodChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void ToolsWidget::setScattererEnabled(QString const & _t1, bool _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2);
}

// SIGNAL 3
void ToolsWidget::reloadShadersClicked()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void ToolsWidget::setFlatSolarSpectrum()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void ToolsWidget::resetSolarSpectrum()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void ToolsWidget::setBlackBodySolarSpectrum(double _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void ToolsWidget::windowDecorationToggled(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void ToolsWidget::projectionChanged(GLWidget::Projection _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}

// SIGNAL 9
void ToolsWidget::colorModeChanged(GLWidget::ColorMode _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1);
}
QT_WARNING_POP
